# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Results for the users service."""

from invenio_records_resources.services.records.results import RecordItem, RecordList


class UserItem(RecordItem):
    """Single user result."""

    def __init__(
        self,
        service,
        identity,
        user,
        errors=None,
        links_tpl=None,
        schema=None,
        **kwargs,
    ):
        """Constructor."""
        self._data = None
        self._errors = errors
        self._identity = identity
        self._user = user
        self._service = service
        self._links_tpl = links_tpl
        self._schema = schema or service.schema

    @property
    def id(self):
        """Identity of the user."""
        return str(self._user.id)

    def __getitem__(self, key):
        """Key a key from the data."""
        return self.data[key]

    @property
    def links(self):
        """Get links for this result item."""
        return self._links_tpl.expand(self._identity, self._user)

    @property
    def _obj(self):
        """Return the object to dump."""
        return self._user

    @property
    def data(self):
        """Property to get the user."""
        if self._data:
            return self._data

        self._data = self._schema.dump(
            self._obj,
            context={
                "identity": self._identity,
                "record": self._user,
            },
        )

        if self._links_tpl:
            self._data["links"] = self.links

        return self._data

    @property
    def errors(self):
        """Get the errors."""
        return self._errors

    def to_dict(self):
        """Get a dictionary for the user."""
        res = self.data
        if self._errors:
            res["errors"] = self._errors
        return res


class UserList(RecordList):
    """List of user results."""

    @property
    def hits(self):
        """Iterator over the hits."""
        user_cls = self._service.record_cls

        for hit in self._results:
            # load dump
            user = user_cls.loads(hit.to_dict())
            schema = self._service.schema

            # project the user
            projection = schema.dump(
                user,
                context={
                    "identity": self._identity,
                    "record": user,
                },
            )

            # inject the links
            if self._links_item_tpl:
                projection["links"] = self._links_item_tpl.expand(self._identity, user)

            yield projection

    def to_dict(self):
        """Return result as a dictionary."""
        # TODO: This part should imitate the result item above. I.e. add a
        # "data" property which uses a ServiceSchema to dump the entire object.
        res = {
            "hits": {
                "hits": list(self.hits),
                "total": self.total,
            }
        }

        if self.aggregations:
            res["aggregations"] = self.aggregations

        if self._params:
            res["sortBy"] = self._params["sort"]
            if self._links_tpl:
                res["links"] = self._links_tpl.expand(self._identity, self.pagination)

        return res



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2024 CERN.
# Copyright (C) 2023 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Domains service configuration."""

from invenio_i18n import lazy_gettext as _
from invenio_records_resources.services import (
    RecordServiceConfig,
    SearchOptions,
    pagination_links,
)
from invenio_records_resources.services.base.config import (
    ConfiguratorMixin,
    FromConfigSearchOptions,
    SearchOptionsMixin,
)
from invenio_records_resources.services.records.params import (
    FacetsParam,
    PaginationParam,
    QueryStrParam,
    SortParam,
)
from invenio_records_resources.services.records.queryparser import QueryParser

from ...records.api import DomainAggregate
from ..common import Link
from ..permissions import DomainPermissionPolicy
from ..schemas import DomainSchema
from .components import DomainComponent


class DomainsSearchOptions(SearchOptions, SearchOptionsMixin):
    """Search options."""

    pagination_options = {
        "default_results_per_page": 30,
        "default_max_results": 1000,
    }

    query_parser_cls = QueryParser.factory(
        fields=[
            "id",
            "domain^3",
        ],
    )

    sort_default = "bestmatch"
    sort_default_no_query = "newest"

    params_interpreters_cls = [
        QueryStrParam,
        SortParam,
        PaginationParam,
        FacetsParam,
    ]


def domainvar(obj, vars):
    """Add domain into link vars."""
    vars["domain"] = obj.domain


class DomainsServiceConfig(RecordServiceConfig, ConfiguratorMixin):
    """Requests service configuration."""

    # common configuration
    permission_policy_cls = DomainPermissionPolicy
    search = FromConfigSearchOptions(
        "USERS_RESOURCES_DOMAINS_SEARCH",
        "USERS_RESOURCES_DOMAINS_SORT_OPTIONS",
        "USERS_RESOURCES_DOMAINS_SEARCH_FACETS",
        search_option_cls=DomainsSearchOptions,
    )

    # specific configuration
    service_id = "domains"
    record_cls = DomainAggregate
    schema = DomainSchema
    indexer_queue_name = "domains"
    index_dumper = None

    # links configuration
    links_item = {
        "self": Link("{+api}/domains/{domain}", vars=domainvar),
        "admin_self_html": Link(
            "{+ui}/administration/domains/{domain}", vars=domainvar
        ),
        "admin_users_html": Link(
            "{+ui}/administration/users?q=domain:{domain}", vars=domainvar
        ),
    }
    links_search = pagination_links("{+api}/domains{?args*}")

    components = [
        DomainComponent,
    ]



# -*- coding: utf-8 -*-
#
# Copyright (C) 2024 CERN
# Copyright (C) 2024 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Data-layer definitions for user and group management in Invenio."""

from invenio_accounts.models import DomainOrg, UserIdentity
from invenio_accounts.proxies import current_datastore
from invenio_accounts.utils import DomainStatus
from invenio_db import db
from invenio_records_resources.records.systemfields.calculated import CalculatedField


class CalculatedIndexedField(CalculatedField):
    """Field that also indexes it's calculated value."""

    def __init__(self, key=None, use_cache=False, index=False):
        """Constructor."""
        super().__init__(key, use_cache=use_cache)
        self._index = index

    def pre_dump(self, record, data, dumper=None):
        """Called after a record is dumped."""
        if self._index:
            data[self.attr_name] = self.obj(record)

    def post_load(self, record, data, loader=None):
        """Called after a record is loaded."""
        if self._index:
            value = data.pop(self.attr_name, None)
            # Store on cache so if cache is used we don't fetch the object again.
            self._set_cache(record, value)


class AccountStatusField(CalculatedIndexedField):
    """Dump a combined account status value."""

    def calculate(self, user_record):
        """Logic for calculating the record's property."""
        status = "new"
        if user_record.active:
            if user_record.confirmed_at and user_record.verified_at:
                status = "verified"
            elif user_record.confirmed_at:
                status = "confirmed"
        else:
            if user_record.blocked_at:
                status = "blocked"
            else:
                status = "inactive"
        return status


class AccountVisibilityField(CalculatedIndexedField):
    """Dump a combined visibility status value."""

    def calculate(self, user_record):
        """Logic for calculating visibility status."""
        if user_record.preferences["email_visibility"] == "public":
            return "full"
        elif user_record.preferences["visibility"] == "public":
            return "profile"
        else:
            return "hidden"


class IsNotNoneField(CalculatedIndexedField):
    """Dump a bool for easier checking if a value is set."""

    def __init__(self, field, *args, **kwargs):
        """Constructor."""
        self._field = field
        super().__init__(*args, **kwargs)

    def calculate(self, user_record):
        """Checks if a timestamp is not none."""
        return getattr(user_record, self._field, None) is not None


class DomainField(CalculatedIndexedField):
    """Get information about the user's domain."""

    def calculate(self, user_record):
        """Checks if a timestamp is not none."""
        domain = current_datastore.find_domain(user_record.domain)
        if domain is None:
            return {
                "tld": "",
                "status": "1",
                "category": None,
                "flagged": False,
            }
        else:
            return {
                "tld": domain.tld,
                "status": domain.status.value,
                "category": domain.category,
                "flagged": domain.flagged,
            }


class UserIdentitiesField(CalculatedIndexedField):
    """Get a user's different linked account identities."""

    def calculate(self, user_record):
        """Checks if a timestamp is not none."""
        identities = (
            db.session.query(UserIdentity).filter_by(id_user=user_record.id).all()
        )
        data = {i.method: i.id for i in identities}
        return data


class DomainOrgField(CalculatedIndexedField):
    """Get information about the user's domain."""

    def calculate(self, domain_record):
        """Checks if a timestamp is not none."""
        if not domain_record.model.org_id:
            return None

        org = domain_record.model.model_obj.org

        parent_org = None
        if org.parent_id is not None:
            parent_org = org.parent

        data = [
            {
                "id": org.id,
                "pid": org.pid,
                "name": org.name,
                "props": org.json or {},
                "is_parent": False,
            }
        ]
        if parent_org:
            data.append(
                {
                    "id": parent_org.id,
                    "pid": parent_org.pid,
                    "name": parent_org.name,
                    "props": parent_org.json or {},
                    "is_parent": True,
                }
            )
        return data


class DomainCategoryNameField(CalculatedIndexedField):
    """Dump the name of the category."""

    def calculate(self, domain_record):
        """Logic for calculating."""
        if domain_record.model.model_obj.category:
            return domain_record.model.model_obj.category_name.label
        else:
            return None


class DomainStatusNameField(CalculatedIndexedField):
    """Dump the name of the category."""

    def calculate(self, domain_record):
        """Logic for calculating."""
        return DomainStatus(domain_record.status).name



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 CERN.
# Copyright (C) 2022 TU Wien.
# Copyright (C) 2023-2024 Graz University of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Invenio module providing management APIs for users and roles/groups."""

from importlib_metadata import entry_points
from invenio_accounts.proxies import current_db_change_history
from invenio_accounts.signals import datastore_post_commit, datastore_pre_commit
from invenio_db import db
from sqlalchemy import event

from . import config
from .records.hooks import post_commit, pre_commit
from .resources import (
    DomainsResource,
    DomainsResourceConfig,
    GroupsResource,
    GroupsResourceConfig,
    UsersResource,
    UsersResourceConfig,
)
from .services import (
    DomainsService,
    DomainsServiceConfig,
    GroupsService,
    GroupsServiceConfig,
    UsersService,
    UsersServiceConfig,
)


class InvenioUsersResources(object):
    """Invenio-Users-Resources extension."""

    def __init__(self, app=None):
        """Extension initialization."""
        if app:
            self.init_app(app)

    def init_app(self, app):
        """Flask application initialization."""
        self.init_config(app)
        self.init_services(app)
        self.init_resources(app)
        self.init_db_hooks()
        self.init_actions_registry()
        app.extensions["invenio-users-resources"] = self

    def init_config(self, app):
        """Initialize configuration."""
        for k in dir(config):
            if k.startswith("USERS_RESOURCES_"):
                app.config.setdefault(k, getattr(config, k))

    def init_services(self, app):
        """Initialize the services for users and user groups."""
        self.users_service = UsersService(config=UsersServiceConfig.build(app))
        self.groups_service = GroupsService(config=GroupsServiceConfig)
        self.domains_service = DomainsService(config=DomainsServiceConfig.build(app))

    def init_resources(self, app):
        """Initialize the resources for users and user groups."""
        self.users_resource = UsersResource(
            service=self.users_service,
            config=UsersResourceConfig,
        )
        self.groups_resource = GroupsResource(
            service=self.groups_service,
            config=GroupsResourceConfig,
        )

        self.domains_resource = DomainsResource(
            service=self.domains_service,
            config=DomainsResourceConfig,
        )

    def init_db_hooks(self):
        """Initialize the database hooks for reindexing updated users/roles."""
        # make sure that the hooks are only registered once per DB connection
        # and not once per app
        if hasattr(db, "_user_resources_hooks_registered"):
            return
        db._user_resources_hooks_registered = True
        datastore_pre_commit.connect(pre_commit)
        datastore_post_commit.connect(post_commit)

        @event.listens_for(db.session, "after_rollback")
        def _after_rollback(session):
            """When a session is rolled back, we don't reindex anything."""
            # this event listener is triggered when the entire transaction is
            # rollbacked. In a case (e.g. nested transactions) when an
            # exception is caught and then the session is commited,
            # the sets might not reflect all the users that were changed.
            current_db_change_history.clear_dirty_sets(session)

    def init_actions_registry(self):
        """Initialises moderation actions registry."""
        self.actions_registry = {}
        self._register_entry_point(
            self.actions_registry,
            "invenio_users_resources.moderation.actions",
        )

    def _register_entry_point(self, registry, ep_name):
        """Load entry points into the given registry."""
        for ep in set(entry_points(group=ep_name)):
            # Entry point has the action as the name (e.g. invenio_users_resources.moderation.actions.block = ... , 'block' is the name)
            action_name = ep.name
            action = ep.load()
            assert callable(action)
            registry.setdefault(action_name, []).append(action)


def finalize_app(app):
    """Finalize app.

    NOTE: replace former @record_once decorator
    """
    init(app)


def api_finalize_app(app):
    """Finalize app for api.

    NOTE: replace former @record_once decorator
    """
    init(app)


def init(app):
    """Init app.

    Register services - cannot be done in extension because
    Invenio-Records-Resources might not have been initialized.
    """
    rr_ext = app.extensions["invenio-records-resources"]
    ext = app.extensions["invenio-users-resources"]
    idx_ext = app.extensions["invenio-indexer"]

    # services
    rr_ext.registry.register(ext.users_service)
    rr_ext.registry.register(ext.groups_service)
    rr_ext.registry.register(ext.domains_service)

    idx_ext.registry.register(ext.users_service.indexer, indexer_id="users")
    idx_ext.registry.register(ext.groups_service.indexer, indexer_id="groups")
    idx_ext.registry.register(ext.domains_service.indexer, indexer_id="domains")



# -*- coding: utf-8 -*-
#
# Copyright (C) 2022 European Union.
# Copyright (C) 2022 CERN.
# Copyright (C) 2024 KTH Royal Institute of Technology.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""User resource tests."""

import pytest


#
# Read
#
def test_read_self_serialization(client, headers, users, user_pub):
    """Read self user."""
    client = user_pub.login(client)
    r = client.get(f"/users/{user_pub.id}", headers=headers)
    assert r.status_code == 200
    data = r.json
    assert data["id"] == user_pub.id
    assert isinstance(data["id"], str)  # must be a string and not an integer
    assert data["email"] == user_pub.email
    assert data["username"] == user_pub.username
    assert data["is_current_user"] is True
    assert data["active"] is True
    assert data["confirmed"] is True
    assert data["profile"] == {
        "full_name": "Jose Benito Gonzalez Lopez",
        "affiliations": "CERN",
    }

    assert data["preferences"] == {
        "email_visibility": "public",
        "visibility": "public",
        "timezone": "Europe/Zurich",
        "locale": "en",
    }
    assert "created" in data
    assert "updated" in data
    assert "domain" in data
    assert "revision_id" in data

    assert "blocked_at" not in data
    assert "confirmed_at" not in data
    assert "current_login_at" not in data
    assert "domaininfo" not in data
    assert "status" not in data
    assert "verified_at" not in data
    assert "verified_at" not in data
    assert "visibility" not in data

    assert data["links"] == {
        "self": f"https://127.0.0.1:5000/api/users/{user_pub.id}",
        "avatar": f"https://127.0.0.1:5000/api/users/{user_pub.id}/avatar.svg",
        "records_html": f"https://127.0.0.1:5000/search/records?q=parent.access.owned_by.user:{user_pub.id}",
    }


@pytest.mark.parametrize(
    "username,public_email",
    [("pub", True), ("pubres", False)],
)
def test_read_anon_serialization(client, headers, users, username, public_email):
    """Read public user as anon."""
    u = users[username]
    # No login
    r = client.get(f"/users/{u.id}", headers=headers)
    assert r.status_code == 200

    data = r.json
    assert data["id"] == u.id
    assert isinstance(data["id"], str)  # must be a string and not an integer
    if public_email:
        assert data["email"] == u.email
    else:
        assert "email" not in data
    assert data["username"] == u.username
    assert data["is_current_user"] is False
    assert data["profile"] == {
        "full_name": u.user.user_profile["full_name"],
        "affiliations": u.user.user_profile["affiliations"],
    }
    assert data["links"] == {
        "self": f"https://127.0.0.1:5000/api/users/{u.id}",
        "avatar": f"https://127.0.0.1:5000/api/users/{u.id}/avatar.svg",
        "records_html": f"https://127.0.0.1:5000/search/records?q=parent.access.owned_by.user:{u.id}",
    }

    for k in [
        "active",
        "confirmed",
        "preferences",
        "created",
        "updated",
        "revision_id",
    ]:
        assert k not in data


#
# Avatar
#
def test_user_avatar(client, user_pub):
    res = client.get(f"/users/{user_pub.id}/avatar.svg")
    assert res.status_code == 200
    assert res.mimetype == "image/svg+xml"
    assert res.headers["Etag"].startswith('"J')  # (J)ose Benito Gonzalez Lopez
    data = res.get_data()


def test_user_avatar_non_ascii(client, user_accented):
    client = user_accented.login(client)
    res = client.get(f"/users/{user_accented.id}/avatar.svg")
    assert res.status_code == 200
    assert res.mimetype == "image/svg+xml"
    assert res.headers["Etag"].startswith('"C')  # (Č)estmír Kopecký
    data = res.get_data()


#
# Management / moderation
#


def test_approve_user(client, headers, user_pub, user_moderator, db):
    """Tests approve user endpoint."""
    client = user_moderator.login(client)
    res = client.post(f"/users/{user_pub.id}/approve", headers=headers)
    assert res.status_code == 200

    res = client.get(f"/users/{user_pub.id}")
    assert res.status_code == 200
    assert res.json["verified_at"] is not None

    # Test user tries to approve themselves
    res = client.post(f"/users/{user_moderator.id}/approve", headers=headers)
    assert res.status_code == 403


def test_block_user(client, headers, user_pub, user_moderator, db):
    """Tests block user endpoint."""
    client = user_moderator.login(client)
    res = client.post(f"/users/{user_pub.id}/block", headers=headers)
    assert res.status_code == 200

    res = client.get(f"/users/{user_pub.id}")
    assert res.status_code == 200
    assert res.json["blocked_at"] is not None

    # Test user tries to block themselves
    res = client.post(f"/users/{user_moderator.id}/block", headers=headers)
    assert res.status_code == 403

    res = client.get(f"/users/{user_moderator.id}")
    assert res.status_code == 200


def test_deactivate_user(client, headers, user_pub, user_moderator, db):
    """Tests deactivate user endpoint."""
    client = user_moderator.login(client)
    res = client.post(f"/users/{user_pub.id}/deactivate", headers=headers)
    assert res.status_code == 200

    res = client.get(f"/users/{user_pub.id}")
    assert res.status_code == 200
    assert res.json["active"] == False

    # Test user tries to deactivate themselves
    res = client.post(f"/users/{user_moderator.id}/deactivate", headers=headers)
    assert res.status_code == 403

    res = client.get(f"/users/{user_moderator.id}")
    assert res.status_code == 200


def test_management_permissions(client, headers, user_pub, db):
    """Test permissions at the resource level."""
    client = user_pub.login(client)
    res = client.post(f"/users/{user_pub.id}/deactivate", headers=headers)
    assert res.status_code == 403


def test_impersonate_user(client, headers, user_pub, user_moderator, db):
    """Tests user impersonation endpoint."""
    client = user_moderator.login(client)
    res = client.get(f"/users/{user_moderator.id}")
    assert res.status_code == 200
    assert res.json["is_current_user"] == True
    res = client.get(f"/users/{user_pub.id}")
    assert res.status_code == 200
    assert res.json["is_current_user"] == False

    res = client.post(f"/users/{user_pub.id}/impersonate", headers=headers)
    assert res.status_code == 200

    res = client.get(f"/users/{user_pub.id}")
    assert res.status_code == 200
    assert res.json["is_current_user"] == True

    res = client.get(f"/users/{user_moderator.id}")
    assert res.status_code == 403

    res = client.post(f"/users/{user_moderator.id}/impersonate", headers=headers)
    assert res.status_code == 403
    res = client.post(f"/users/{user_pub.id}/impersonate", headers=headers)
    assert res.status_code == 403


@pytest.mark.parametrize(
    "link_name,expected_url",
    [
        (
            "admin_records_html",
            "/administration/records?q=parent.access.owned_by.user:{id}&f=allversions",
        ),
        (
            "admin_drafts_html",
            "/administration/drafts?q=parent.access.owned_by.user:{id}&f=allversions",
        ),
        ("admin_moderation_html", "/administration/moderation?q=topic.user:{id}"),
    ],
)
def test_admin_links(
    client, headers, user_moderator, user_pub, link_name, expected_url
):
    """Test admin links."""
    client = user_moderator.login(client)
    res = client.get(f"/users/{user_pub.id}", headers=headers)
    assert res.status_code == 200
    data = res.json
    assert link_name in data["links"]
    assert (
        data["links"][link_name]
        == f"https://127.0.0.1:5000{expected_url.format(id=user_pub.id)}"
    )


@pytest.mark.parametrize(
    "username,expected_admin_links",
    [
        ("user_moderator", True),  # user_moderator should have admin links
        ("pub", False),  # regular user should not have admin links
        ("res", False),  # regular user should not have admin links
    ],
)
def test_admin_links_visibility(client, headers, users, username, expected_admin_links):
    """Test admin links visibility based on user permissions."""
    user = users[username]
    client = user.login(client)
    res = client.get(f"/users/{user.id}", headers=headers)
    assert res.status_code == 200
    data = res.json

    if expected_admin_links:
        assert "admin_records_html" in data["links"]
        assert "admin_drafts_html" in data["links"]
        assert "admin_moderation_html" in data["links"]
    else:
        assert "admin_records_html" not in data["links"]
        assert "admin_drafts_html" not in data["links"]
        assert "admin_moderation_html" not in data["links"]


# TODO: test conditional requests
# TODO: test caching headers
# TODO: test invalid identifiers
# TODO: test concurrency for management actions (block then restore and vice-versa)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2025 Ubiquity Press.
#
# Invenio-Users-Resources is free software; you can redistribute it and/or
# modify it under the terms of the MIT License; see LICENSE file for more
# details.

"""Permission generators tests."""

from invenio_access.utils import get_identity
from invenio_records_permissions.generators import AuthenticatedUser

from invenio_users_resources.permissions import user_management_action
from invenio_users_resources.services.generators import IfGroupNotManaged
from invenio_users_resources.services.permissions import UserManager


def test_group_not_managed_generator(app, user_pub, user_moderator):
    """Test IfGroupNotManaged generator."""

    permission = IfGroupNotManaged([AuthenticatedUser()], [UserManager])

    assert permission.needs() == {user_management_action}
    assert permission.needs(record={"is_managed": True}) == {user_management_action}

    identity = get_identity(user_pub)
    query = permission.query_filter(identity=identity)
    assert query.to_dict() == {"match": {"is_managed": False}}

    identity = get_identity(user_moderator)
    query = permission.query_filter(identity=identity)
    assert query.to_dict() == {"match_all": {}}
